import copy
import logging
import math
import time
import hashlib
import random
import numpy as np

logger = logging.getLogger('graphsignal')


RESERVOIR_SIZE = 100
HISTOGRAM_BIN_COUNT = 10


class Window(object):
    __slots__ = [
        'model',
        'metrics',
        'samples',
        'events',
        'timestamp'
    ]

    def __init__(
            self,
            model=None,
            timestamp=None):
        self.model = model
        self.metrics = None
        self.samples = None
        self.events = None
        self.timestamp = timestamp if timestamp is not None else _now()

    def add_metric(self, metric):
        if self.metrics is None:
            self.metrics = []
        self.metrics.append(metric)

    def add_sample(self, sample):
        if self.samples is None:
            self.samples = []
        self.samples.append(sample)

    def add_event(self, event):
        if self.events is None:
            self.events = []
        self.events.append(event)

    def to_dict(self):
        metric_dicts = [metric.to_dict()
                        for metric in self.metrics] if self.metrics else None
        sample_dicts = [sample.to_dict()
                        for sample in self.samples] if self.samples else None
        event_dicts = [event.to_dict()
                       for event in self.events] if self.events else None

        window_dict = {
            'model': self.model.to_dict(),
            'metrics': metric_dicts,
            'samples': sample_dicts,
            'events': event_dicts,
            'timestamp': self.timestamp
        }

        for key in list(window_dict):
            if window_dict[key] is None:
                del window_dict[key]

        return window_dict

    def __str__(self):
        report = []

        report.append('Model')
        if self.model:
            report.append('    name: {0}'.format(self.model.name))
            report.append('    deployment: {0}'.format(self.model.deployment))
            if self.model.attributes is not None:
                report.append(
                    '    attributes: {0}'.format(
                        self.model.attributes))

        if self.metrics is not None:
            report.append('Metrics ({0})'.format(len(self.metrics)))
            for metric in self.metrics:
                report.append('    {0}:{1}:{2}: {3} (type: {4})'.format(
                    metric.dataset, metric.dimension, metric.name, metric.measurement, metric.type))

        if self.events is not None:
            report.append('Events ({0})'.format(len(self.events)))
            for event in self.events:
                report.append('    {0}: {1}'.format(
                    event.type, event.description))
                if event.attributes is not None:
                    for attribute in event.attributes:
                        report.append(
                            '        attribute: {0}: {1}'.format(attribute.name, attribute.value))

        if self.samples is not None:
            report.append('Samples ({0})'.format(len(self.samples)))
            for sample in self.samples:
                report.append('    {0}'.format(sample.name))
                for segment in sample.segments:
                    report.append(
                        '        dataset: {0}'.format(
                            segment.dataset))
                    report.append('        format: {0}'.format(segment.format))
                    report.append(
                        '        data size: {0}'.format(len(segment.data)))

        return '\n'.join(report)


class Attribute(object):
    __slots__ = ['name', 'value']

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def to_dict(self):
        return {
            'name': self.name,
            'value': self.value
        }


class Model(object):
    __slots__ = [
        'name',
        'deployment',
        'attributes',
        'timestamp'
    ]

    def __init__(
            self,
            name=None,
            deployment=None,
            timestamp=None):
        self.name = name
        self.deployment = deployment
        self.attributes = None
        self.timestamp = timestamp if timestamp is not None else _now()

    def add_attribute(self, name, value):
        if self.attributes is None:
            self.attributes = []
        self.attributes.append(Attribute(name, str(value)))

    def to_dict(self):
        attribute_dicts = [attribute.to_dict(
        ) for attribute in self.attributes] if self.attributes else None

        model_dict = {
            'name': self.name,
            'deployment': self.deployment,
            'attributes': attribute_dicts,
            'timestamp': self.timestamp
        }

        for key in list(model_dict):
            if model_dict[key] is None:
                del model_dict[key]

        return model_dict


class Metric(object):
    __slots__ = [
        'dataset',
        'dimension',
        'name',
        'type',
        'unit',
        'measurement',
        'timestamp',
        '_reservoir',
        '_percent'
    ]

    DATASET_INPUT = 'input'
    DATASET_OUTPUT = 'output'
    DATASET_SYSTEM = 'system'
    DATASET_USER_DEFINED = 'user_defined'

    TYPE_GAUGE = 'gauge'
    TYPE_STATISTIC = 'statistic'
    TYPE_HISTOGRAM = 'histogram'

    UNIT_NONE = ''
    UNIT_PERCENT = '%'
    UNIT_MILLISECOND = 'ms'
    UNIT_KILOBYTE = 'KB'

    def __init__(
            self,
            dataset=None,
            dimension=None,
            name=None,
            timestamp=None):
        self.dataset = dataset
        self.dimension = dimension
        self.name = name
        self.type = None
        self.unit = None
        self.measurement = None
        self.timestamp = timestamp if timestamp is not None else _now()
        self._reservoir = None
        self._percent = None

    def set_gauge(self, value, unit=UNIT_NONE):
        self.type = Metric.TYPE_GAUGE
        self.measurement = [value]
        self.unit = unit

    def set_statistic(self, statistic, sample_size, unit=UNIT_NONE):
        self.type = Metric.TYPE_STATISTIC
        self.measurement = [statistic, sample_size]
        self.unit = unit

    def update_percentile(self, value, percent, unit=UNIT_NONE):
        if not self.type:
            self.type = Metric.TYPE_STATISTIC
            self.unit = unit
            self.measurement = [None, 0]
            self._reservoir = []
            self._percent = percent

        if len(self._reservoir) < RESERVOIR_SIZE:
            self._reservoir.append(value)
        else:
            self._reservoir[random.randint(0, RESERVOIR_SIZE - 1)] = value

        self.measurement[1] += 1

    def compute_histogram(self, values, unit=UNIT_NONE):
        if len(values) == 0:
            return None

        self.type = Metric.TYPE_HISTOGRAM
        self.unit = unit
        self.measurement = []

        m_max = max(values)
        m_range = m_max - min(values)

        if m_range == 0:
            self.measurement.append(0)
            self.measurement.extend([values[0], len(values)])
            return

        bin_size = 10 ** math.floor(math.log10(m_range /
                                               float(HISTOGRAM_BIN_COUNT)))

        bin_exp = math.floor(math.log10(bin_size))
        if bin_exp > 0:
            bin_exp = 0
        else:
            bin_exp = int(abs(bin_exp))

        hist = {}
        for m in values:
            bin = None
            if bin_size > 0:
                bin = math.floor(m / bin_size) * bin_size
            else:
                bin = m

            if bin in hist:
                hist[bin] += 1
            else:
                hist[bin] = 1

        self.measurement.append(bin_size)
        for bin in sorted(hist):
            self.measurement.append(round(bin, bin_exp))
            self.measurement.append(hist[bin])

    def compute_categorical_histogram(self, values, unit=UNIT_NONE):
        if len(values) == 0:
            return None

        self.type = Metric.TYPE_HISTOGRAM
        self.unit = unit
        self.measurement = []

        n_components = 100
        counts = {}
        for value in values:
            h = _category_hash(value, n_components=n_components)
            if h in counts:
                counts[h] += 1
            else:
                counts[h] = 1

        self.measurement = [1]
        for i in range(n_components):
            if i in counts:
                self.measurement.append(i)
                self.measurement.append(counts[i])

    def finalize(self):
        if self._reservoir is not None:
            size = len(self._reservoir)
            index = int(math.ceil((size * self._percent) / 100)) - 1
            self.measurement[0] = sorted(self._reservoir)[index]

    def to_dict(self):
        metric_dict = {
            'dataset': self.dataset,
            'dimension': self.dimension,
            'name': self.name,
            'type': self.type,
            'unit': self.unit,
            'measurement': self.measurement,
            'timestamp': self.timestamp
        }

        for key in list(metric_dict):
            if metric_dict[key] is None:
                del metric_dict[key]

        return metric_dict


def _category_hash(category, n_components=256):
    # hashing trick
    h = hashlib.md5(category.encode('utf-8')).hexdigest()
    return int(h, 16) % n_components


class Segment(object):
    __slots__ = [
        'dataset',
        'format',
        'data']

    def __init__(self, dataset, format_, data):
        self.dataset = dataset
        self.format = format_
        self.data = data

    def to_dict(self):
        return {
            'dataset': self.dataset,
            'format': self.format,
            'data': self.data
        }


class Sample(object):
    __slots__ = [
        'name',
        'size',
        'segments',
        'timestamp'
    ]

    SEGMENT_FORMAT_CSV = 'csv'

    def __init__(self, name=None, size=None, timestamp=None):
        self.name = name
        self.size = size
        self.segments = []
        self.timestamp = timestamp if timestamp is not None else _now()

    def set_size(self, size):
        self.size = size

    def add_segment(self, dataset, format_, data, insert_at=None):
        if insert_at:
            self.segments.insert(insert_at, Segment(dataset, format_, data))
        else:
            self.segments.append(Segment(dataset, format_, data))

    def to_dict(self):
        segment_dicts = [segment.to_dict()
                         for segment in self.segments] if self.segments else None

        sample_dict = {
            'name': self.name,
            'size': self.size,
            'segments': segment_dicts,
            'timestamp': self.timestamp
        }

        for key in list(sample_dict):
            if sample_dict[key] is None:
                del sample_dict[key]

        return sample_dict


class Event(object):
    __slots__ = [
        'type',
        'score',
        'name',
        'description',
        'attributes',
        'timestamp'
    ]

    TYPE_INFO = 'info'
    TYPE_ERROR = 'error'

    def __init__(
            self,
            type=None,
            score=None,
            name=None,
            description=None,
            timestamp=None):
        self.type = type
        self.score = score if score is not None else 1
        self.name = name
        self.description = description
        self.attributes = None
        self.timestamp = timestamp if timestamp is not None else _now()

    def add_attribute(self, name, value):
        if self.attributes is None:
            self.attributes = []
        self.attributes.append(Attribute(name, str(value)))

    def to_dict(self):
        attribute_dicts = [attribute.to_dict(
        ) for attribute in self.attributes] if self.attributes else None

        event_dict = {
            'type': self.type,
            'score': self.score,
            'name': self.name,
            'description': self.description,
            'attributes': attribute_dicts,
            'timestamp': self.timestamp
        }

        for key in list(event_dict):
            if event_dict[key] is None:
                del event_dict[key]

        return event_dict


def _now():
    return int(time.time())



import re
import os
try:
    import resource
except ImportError:
    resource = None


VM_RSS_REGEXP = re.compile(r'VmRSS:\s+(\d+)\s+kB')
VM_SIZE_REGEXP = re.compile(r'VmSize:\s+(\d+)\s+kB')


def cpu_time():
    if resource:
        rusage = resource.getrusage(resource.RUSAGE_SELF)
        return rusage.ru_utime + rusage.ru_stime

    return None


def vm_rss():
    output = None
    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()
    except Exception:
        return None

    match = VM_RSS_REGEXP.search(output)
    if match:
        return int(float(match.group(1)))

    return None


def vm_size():
    output = None
    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()
    except Exception:
        return None

    match = VM_SIZE_REGEXP.search(output)
    if match:
        return int(float(match.group(1)))

    return None



import unittest
import logging
from unittest.mock import patch, Mock

import graphsignal

logger = logging.getLogger('graphsignal')


class GraphsignalTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(api_key='k1', debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_configure(self):
        self.assertEqual(graphsignal._get_config().api_key, 'k1')
        self.assertEqual(graphsignal._get_config().debug_mode, True)
        self.assertEqual(graphsignal._get_config().log_instances, True)



import logging
import time

from graphsignal.windows import Metric
from graphsignal import system

logger = logging.getLogger('graphsignal')


class Span(object):
    __slots__ = [
        '_session',
        '_start_time',
        '_start_cpu_time',
        '_stopped'
    ]

    def __init__(self, session):
        self._session = session
        self._start_time = None
        self._stopped = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

        if exc_type and exc_val and exc_tb:
            self._session.log_event(
                description='Prediction exception',
                exc_info=(exc_type, exc_val, exc_tb)
            )

    def start(self):
        if self._stopped:
            return
        self._start_time = time.monotonic()
        self._start_cpu_time = system.cpu_time()

    def stop(self):
        if self._stopped:
            return
        self._stopped = True

        self._set_p50_metric(
            'prediction_latency_p50',
            round((time.monotonic() - self._start_time) * 1000))

        if self._start_cpu_time:
            self._set_p50_metric(
                'prediction_cpu_time_p50',
                round((system.cpu_time() - self._start_cpu_time) * 1000))

        self._session._set_updated()

    def _set_p50_metric(self, name, value):
        with self._session._update_lock:
            metric = None
            if name in self._session._metric_index:
                metric = self._session._metric_index[name]
            else:
                metric = self._session._metric_index[name] = Metric(
                    dataset=Metric.DATASET_SYSTEM,
                    name=name)
            metric.update_percentile(value, 50, Metric.UNIT_MILLISECOND)



import unittest
import logging


from unittest.mock import patch, Mock

from graphsignal import system

logger = logging.getLogger('graphsignal')


class SystemTest(unittest.TestCase):
    def setUp(self):
        logger.setLevel(logging.DEBUG)

    def tearDown(self):
        pass

    def test_cpu_time(self):
        self.assertTrue(system.cpu_time() > 0)

    def test_vm_rss(self):
        self.assertTrue(system.vm_rss() > 0)

    def test_vm_size(self):
        self.assertTrue(system.vm_size() > 0)



import re
import unicodedata
import time
import logging
import numpy as np
import pandas as pd
from scipy.stats import skew, kurtosis
from sklearn.neighbors import LocalOutlierFactor

import graphsignal
from graphsignal.windows import Metric, Sample

logger = logging.getLogger('graphsignal')
_rand = np.random.RandomState(int(time.time()))

MAX_COLUMNS = 250
RANDOM_SAMPLE_SIZE = 10
OUTLIER_SAMPLE_SIZE = 10
MIN_INSTANCES_FOR_OUTLIER_DETECTION = 30


def estimate_size(data):
    if data is None:
        return 0

    if isinstance(data, list):
        return len(data)
    elif isinstance(data, dict):
        if len(data) > 0:
            elem = next(iter(data.values()))
            if isinstance(elem, list):
                return len(elem)
            elif isinstance(elem, np.ndarray):
                return elem.shape[0] if elem.ndim > 0 else 1
    elif isinstance(data, np.ndarray):
        return data.shape[0] if data.ndim > 0 else 1
    elif isinstance(data, pd.DataFrame):
        return data.shape[0]
    return 0


def compute_metrics(prediction_window):
    start_ts = time.time()
    metrics = []
    samples = []

    if len(prediction_window) == 0:
        return metrics, samples

    last_timestamp = max([p.timestamp for p in prediction_window if p])

    # convert data to 2d
    input_data_window = [(p.input_data, p.timestamp)
                        for p in prediction_window if p.input_data is not None]
    input_data2d, input_timestamps = _convert_window_to_2d(input_data_window)

    output_data_window = [(p.output_data, p.timestamp)
                         for p in prediction_window if p.output_data is not None]
    output_data2d, output_timestamps = _convert_window_to_2d(output_data_window)

    context_data_window = [(p.context_data, p.timestamp)
                          for p in prediction_window if p.context_data is not None]
    context_data2d, _ = _convert_window_to_2d(context_data_window)

    if input_data2d is None and output_data2d is None:
        logger.warning('Provided empty data, nothing to compute')
        return metrics, samples

    # add timestamp to context
    timestamps = input_timestamps
    if input_timestamps is not None:
        timestamps = output_timestamps
    if timestamps is not None:
        if context_data2d is None:
            context_data2d = pd.DataFrame(
                data={'prediction_timestamp': timestamps})
        elif timestamps.shape[0] == context_data2d.shape[0]:
            context_data2d['prediction_timestamp'] = timestamps

    # compute metrics
    if input_data2d is not None:
        metrics.extend(
            _compute_tabular_metrics(
                input_data2d,
                Metric.DATASET_INPUT,
                last_timestamp))

    if output_data2d is not None:
        metrics.extend(
            _compute_tabular_metrics(
                output_data2d,
                Metric.DATASET_OUTPUT,
                last_timestamp))

    # compute samples
    if graphsignal._get_config().log_instances:
        random_sample = _compute_random_sample(
            input_data2d, output_data2d, context_data2d, last_timestamp)
        if random_sample:
            samples.append(random_sample)
        outlier_samples = _compute_outlier_samples(
            input_data2d, output_data2d, context_data2d, last_timestamp)
        if outlier_samples is not None:
            samples.extend(outlier_samples)

    logger.debug('Computing metrics and samples took %.3f sec',
                 time.time() - start_ts)

    return metrics, samples


def _compute_tabular_metrics(data2d, dataset, timestamp):
    metrics = []

    instance_count = data2d.shape[0]

    metric = Metric(
        dataset=dataset,
        name='instance_count',
        timestamp=timestamp)
    metric.set_statistic(instance_count, instance_count)
    metrics.append(metric)

    metric = Metric(
        dataset=dataset,
        name='dimension_count',
        timestamp=timestamp)
    metric.set_statistic(len(data2d.columns), instance_count)
    metrics.append(metric)

    types = data2d.dtypes.tolist()
    columns = _format_names(data2d.columns.values.tolist())

    for column_index in range(min(data2d.shape[1], MAX_COLUMNS)):
        column_name = columns[column_index]
        if column_name.lower() in ('timestamp', 'time', 'datetime', 'date'):
            continue
        column_type = types[column_index]
        if np.issubdtype(column_type, np.datetime64):
            continue
        column_values = data2d[data2d.columns[column_index]].to_numpy()

        if np.issubdtype(column_type, np.number):
            missing_count = np.count_nonzero(np.isnan(column_values))
            missing_count += np.count_nonzero(np.isinf(column_values))
            metric = Metric(
                dataset=dataset,
                dimension=column_name,
                name='missing_values',
                timestamp=timestamp)
            metric.set_statistic(
                missing_count / instance_count * 100,
                instance_count,
                Metric.UNIT_PERCENT)
            metrics.append(metric)

            zero_count = np.count_nonzero(column_values == 0)
            metric = Metric(
                dataset=dataset,
                dimension=column_name,
                name='zero_values',
                timestamp=timestamp)
            metric.set_statistic(
                zero_count / instance_count * 100,
                instance_count,
                Metric.UNIT_PERCENT)
            metrics.append(metric)

            finite_column_values = column_values[np.isfinite(column_values)]
            if finite_column_values.shape[0] > 0:
                unique_count = np.unique(column_values).size
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='unique_values',
                    timestamp=timestamp)
                metric.set_statistic(unique_count, instance_count)
                metrics.append(metric)

                mean_value = np.mean(finite_column_values)
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='mean',
                    timestamp=timestamp)
                metric.set_statistic(mean_value, instance_count)
                metrics.append(metric)

                std_value = np.std(
                    finite_column_values,
                    ddof=min(1, finite_column_values.shape[0] - 1))
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='standard_deviation',
                    timestamp=timestamp)
                metric.set_statistic(std_value, instance_count)
                metrics.append(metric)

                skewness_value = skew(finite_column_values, bias=False)
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='skewness',
                    timestamp=timestamp)
                metric.set_statistic(skewness_value, instance_count)
                metrics.append(metric)

                kurtosis_value = kurtosis(finite_column_values, bias=False)
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='kurtosis',
                    timestamp=timestamp)
                metric.set_statistic(kurtosis_value, instance_count)
                metrics.append(metric)

                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='distribution',
                    timestamp=timestamp)
                metric.compute_histogram(finite_column_values.tolist())
                metrics.append(metric)
        else:
            string_column_values = [
                m for m in column_values.tolist() if isinstance(m, str)]
            if len(string_column_values) > 0:
                missing_count = column_values.shape[0] - \
                    len(string_column_values)
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='missing_values',
                    timestamp=timestamp)
                metric.set_statistic(
                    missing_count / instance_count * 100,
                    instance_count,
                    Metric.UNIT_PERCENT)
                metrics.append(metric)

                empty_count = string_column_values.count('')
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='empty_values',
                    timestamp=timestamp)
                metric.set_statistic(
                    empty_count / instance_count * 100,
                    instance_count,
                    Metric.UNIT_PERCENT)
                metrics.append(metric)

                unique_count = len(set(string_column_values))
                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='unique_values',
                    timestamp=timestamp)
                metric.set_statistic(unique_count, instance_count)
                metrics.append(metric)

                metric = Metric(
                    dataset=dataset,
                    dimension=column_name,
                    name='distribution',
                    timestamp=timestamp)
                metric.compute_categorical_histogram(string_column_values)
                metrics.append(metric)

    return metrics


def _compute_random_sample(
        input_data2d, output_data2d, context_data2d, timestamp):
    sample = Sample(name='random_sample', timestamp=timestamp)
    if input_data2d is not None:
        sample_idx = _random_index(input_data2d)
        sample.set_size(len(sample_idx))
        sample.add_segment(
            dataset='input',
            format_=Sample.SEGMENT_FORMAT_CSV,
            data=_create_csv(
                data=input_data2d.iloc[sample_idx, :].to_numpy().tolist(),
                columns=_format_names(input_data2d.columns.values)))
        if output_data2d is not None and input_data2d.shape[0] == output_data2d.shape[0]:
            sample.add_segment(
                dataset='output',
                format_=Sample.SEGMENT_FORMAT_CSV,
                data=_create_csv(
                    data=output_data2d.iloc[sample_idx, :].to_numpy().tolist(),
                    columns=_format_names(output_data2d.columns.values)))
        if context_data2d is not None and input_data2d.shape[0] == context_data2d.shape[0]:
            sample.add_segment(
                dataset='context',
                format_=Sample.SEGMENT_FORMAT_CSV,
                data=_create_csv(
                    data=context_data2d.iloc[sample_idx,
                                             :].to_numpy().tolist(),
                    columns=_format_names(context_data2d.columns.values)))
    elif output_data2d is not None:
        sample_idx = _random_index(output_data2d)
        sample.set_size(len(sample_idx))
        sample.add_segment(
            dataset='output',
            format_=Sample.SEGMENT_FORMAT_CSV,
            data=_create_csv(
                data=output_data2d.iloc[sample_idx, :].to_numpy().tolist(),
                columns=_format_names(output_data2d.columns.values)))
        if context_data2d is not None and output_data2d.shape[0] == context_data2d.shape[0]:
            sample.add_segment(
                dataset='context',
                format_=Sample.SEGMENT_FORMAT_CSV,
                data=_create_csv(
                    data=context_data2d.iloc[sample_idx,
                                             :].to_numpy().tolist(),
                    columns=_format_names(context_data2d.columns.values)))

    return sample


def _random_index(data2d):
    sample_size = min(data2d.shape[0], RANDOM_SAMPLE_SIZE)
    return _rand.choice(data2d.shape[0], sample_size, replace=False)


def _compute_outlier_samples(
        input_data2d, output_data2d, context_data2d, timestamp):
    samples = []

    if input_data2d is not None:
        sample = Sample(name='input_outliers', timestamp=timestamp)
        sample_idx = _outlier_index(input_data2d)
        if sample_idx is not None and sample_idx.shape[0] > 0:
            sample.set_size(sample_idx.shape[0])
            sample.add_segment(
                dataset='input',
                format_=Sample.SEGMENT_FORMAT_CSV,
                data=_create_csv(
                    data=input_data2d.iloc[sample_idx, :].to_numpy().tolist(),
                    columns=_format_names(input_data2d.columns.values)))
            if output_data2d is not None and input_data2d.shape[0] == output_data2d.shape[0]:
                sample.add_segment(
                    dataset='output',
                    format_=Sample.SEGMENT_FORMAT_CSV,
                    data=_create_csv(
                        data=output_data2d.iloc[sample_idx,
                                                :].to_numpy().tolist(),
                        columns=_format_names(output_data2d.columns.values)))
            if context_data2d is not None and input_data2d.shape[0] == context_data2d.shape[0]:
                sample.add_segment(
                    dataset='context',
                    format_=Sample.SEGMENT_FORMAT_CSV,
                    data=_create_csv(
                        data=context_data2d.iloc[sample_idx,
                                                 :].to_numpy().tolist(),
                        columns=_format_names(context_data2d.columns.values)))
            samples.append(sample)

    if output_data2d is not None:
        sample = Sample(name='output_outliers', timestamp=timestamp)
        sample_idx = _outlier_index(output_data2d)
        if sample_idx is not None and sample_idx.shape[0] > 0:
            sample.set_size(sample_idx.shape[0])
            sample.add_segment(
                dataset='output',
                format_=Sample.SEGMENT_FORMAT_CSV,
                data=_create_csv(
                    data=output_data2d.iloc[sample_idx, :].to_numpy().tolist(),
                    columns=_format_names(output_data2d.columns.values)))
            if input_data2d is not None and output_data2d.shape[0] == input_data2d.shape[0]:
                sample.add_segment(
                    dataset='input',
                    format_=Sample.SEGMENT_FORMAT_CSV,
                    data=_create_csv(
                        data=input_data2d.iloc[sample_idx,
                                               :].to_numpy().tolist(),
                        columns=_format_names(input_data2d.columns.values)),
                    insert_at=0)
            if context_data2d is not None and output_data2d.shape[0] == context_data2d.shape[0]:
                sample.add_segment(
                    dataset='context',
                    format_=Sample.SEGMENT_FORMAT_CSV,
                    data=_create_csv(
                        data=context_data2d.iloc[sample_idx,
                                                 :].to_numpy().tolist(),
                        columns=_format_names(context_data2d.columns.values)))
            samples.append(sample)

    return samples


def _outlier_index(data2d):
    if data2d.shape[0] < MIN_INSTANCES_FOR_OUTLIER_DETECTION:
        return None

    numeric_data2d = data2d.select_dtypes(include=np.number).to_numpy()
    if numeric_data2d.shape[1] == 0:
        return None

    finite_numeric_data2d = numeric_data2d[np.isfinite(
        numeric_data2d).all(axis=1)]
    if finite_numeric_data2d.shape[0] < MIN_INSTANCES_FOR_OUTLIER_DETECTION:
        return None

    try:
        n_neighbors = min(finite_numeric_data2d.shape[0] - 1, 20)
        lof = LocalOutlierFactor(n_neighbors=n_neighbors)
        scores = lof.fit_predict(finite_numeric_data2d)
        factors = lof.negative_outlier_factor_

        top_factor_idx = np.argsort(factors)[0:OUTLIER_SAMPLE_SIZE]
        top_scores = scores[top_factor_idx]
        return top_factor_idx[top_scores == -1]
    except BaseException:
        logger.error('Error detecting outliers', exc_info=True)

    return None


def _create_csv(data, columns):
    rows = []
    rows.append(','.join(_format_values(columns)))
    for instance in data:
        rows.append(','.join(_format_values(instance)))
    return '\n'.join(rows)


def _convert_window_to_2d(data_window):
    if data_window is None:
        return None

    data2d_window = []
    for data, timestamp in data_window:
        data2d = _convert_to_2d(data)
        if data2d is None:
            return None
        data2d_window.append((data2d, np.full((data2d.shape[0],), timestamp)))

    if len(data2d_window) > 0:
        data2d = pd.concat(
            [data2d for d, _ in data2d_window], ignore_index=True)
        timestamps = np.concatenate(
            [timestamps for _, timestamps in data2d_window])
        if not data2d.empty:
            return data2d, timestamps

    return None, None


def _convert_to_2d(data):
    data2d = None

    # list or dict
    if isinstance(data, (list, dict)):
        if _is_1d(data):  # 1-d list or dict
            data2d = pd.DataFrame(data)
        elif _is_2d(data):  # 2-d list or dict
            data2d = pd.DataFrame(data)
        elif isinstance(data, dict):  # N-d dict
            logger.error('Only dict with max 2 dimensions is supported')
            return None
        # list with N-d ndarray
        elif len(data) > 0 and isinstance(data[0], np.ndarray):
            data = np.asarray(data)
            if data.ndim > 2:
                data = np.reshape(data, (-1, data.shape[-1]))
            data2d = pd.DataFrame(data)
        else:  # N-d list of lists and/or dicts
            data2d = pd.DataFrame(_filter_2d(data))

    # numpy ndarray
    elif isinstance(data, np.ndarray):
        if data.ndim > 2:
            data = np.reshape(data, (-1, data.shape[-1]))
        data2d = pd.DataFrame(data)

    # pandas DataFrame
    elif isinstance(data, pd.DataFrame):
        data2d = data.copy(deep=False)

    if data2d is None:
        logger.error(
            'Unsupported data format: {0}. Please use list, dict, numpy.ndarray, pandas.DataFrame'.format(type(data)))

    return data2d


def _is_scalar(data):
    return isinstance(data, (int, float, str))


def _is_1d(data):
    if isinstance(data, list):
        for elem in data:
            if not _is_scalar(elem):
                return False
        return True
    if isinstance(data, dict):
        for elem in data.values():
            if not _is_scalar(elem):
                return False
        return True
    elif isinstance(data, np.ndarray):
        return data.ndim == 1
    return False


def _is_2d(data):
    if isinstance(data, list):
        for elem in data:
            if not _is_1d(elem):
                return False
        return True
    elif isinstance(data, dict):
        for elem in data.values():
            if not _is_1d(elem):
                return False
        return True
    return False


def _filter_2d(data):
    filtered = []
    for instance in data:
        if isinstance(instance, list):
            filtered.append([v for v in instance if _is_scalar(v)])
        elif isinstance(instance, dict):
            filtered.append(
                {k: v for k, v in instance.items() if _is_scalar(v)})
        elif isinstance(instance, (int, float, str)):
            filtered.append([instance])
    return filtered


def _format_names(names):
    return [str(n)[:50] for n in names]


def _format_values(values):
    return [_format_value(value) for value in values]


def _format_value(value):
    if isinstance(value, str):
        return value[0:250]
    else:
        return str(value)



import unittest
import sys
import json
import threading
import time
import gzip
import logging
from io import BytesIO
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.request import urlopen
from urllib.request import Request
from urllib.parse import urlencode
from urllib.error import URLError


from unittest.mock import patch, Mock

import graphsignal
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class UploaderTest(unittest.TestCase):
    def setUp(self):
        logger.setLevel(logging.DEBUG)
        graphsignal._get_uploader().clear()
        graphsignal.configure(api_key='k1', debug_mode=True)

    def tearDown(self):
        graphsignal._get_uploader().clear()
        graphsignal.shutdown()

    @patch.object(Uploader, 'post_json')
    def test_flush(self, mocked_post_json):
        graphsignal._get_uploader().upload_window({'m1': 1})
        graphsignal._get_uploader().flush()

        mocked_post_json.assert_called_once_with('windows', [{'m1': 1}])
        self.assertEqual(len(graphsignal._get_uploader().buffer['windows']), 0)

    @patch.object(Uploader, 'post_json')
    def test_flush_in_thread(self, mocked_post_json):
        graphsignal._get_uploader().upload_window({'m1': 1})
        graphsignal._get_uploader().flush_in_thread()
        graphsignal.tick()

        mocked_post_json.assert_called_once_with('windows', [{'m1': 1}])
        self.assertEqual(len(graphsignal._get_uploader().buffer['windows']), 0)

    @patch.object(Uploader, 'post_json')
    def test_flush_fail(self, mocked_post_json):
        def side_effect(*args):
            raise URLError("Ex1")
        mocked_post_json.side_effect = side_effect

        graphsignal._get_uploader().upload_window({'m1': 1})
        graphsignal._get_uploader().upload_window({'m2': 2})
        graphsignal._get_uploader().flush()

        self.assertEqual(len(graphsignal._get_uploader().buffer['windows']), 2)

    def test_post_json(self):
        graphsignal._get_uploader().collector_url = 'http://localhost:5005'

        server = TestServer(5005)
        server.start()

        graphsignal._get_uploader().post_json('metrics', [{'m1': 1}])

        request_data = json.loads(server.get_request_data())
        self.assertEqual(request_data[0]['m1'], 1)

        server.join()


class TestServer(threading.Thread):
    def __init__(self, port, delay=None, handler_func=None):
        self.port = port
        RequestHandler.delay = delay
        RequestHandler.handler_func = [handler_func]
        RequestHandler.response_data = '{}'
        RequestHandler.response_code = 200
        threading.Thread.__init__(self)
        self.server = HTTPServer(('localhost', self.port), RequestHandler)

    def get_request_data(self):
        return RequestHandler.request_data

    def set_response_data(self, response_data):
        RequestHandler.response_data = response_data

    def set_response_code(self, response_code):
        RequestHandler.response_code = response_code

    def run(self):
        self.server.handle_request()


def timestamp():
    return int(time.time())


class RequestHandler(BaseHTTPRequestHandler):
    delay = None
    handler_func = None
    request_data = None
    response_data = '{}'
    response_code = 200

    def do_GET(self):
        if self.delay:
            time.sleep(self.delay)

        if RequestHandler.handler_func:
            RequestHandler.handler_func[0]()

        self.send_response(RequestHandler.response_code)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(RequestHandler.response_data.encode('utf-8'))

    def do_POST(self):
        if self.delay:
            time.sleep(self.delay)

        self.request_url = self.path
        content_len = int(self.headers.get('content-length'))

        decompressed_data = gzip.GzipFile(
            fileobj=BytesIO(self.rfile.read(content_len))).read()
        RequestHandler.request_data = decompressed_data.decode('utf-8')

        self.send_response(RequestHandler.response_code)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(RequestHandler.response_data.encode('utf-8'))



import logging
import time

logger = logging.getLogger('graphsignal')


class Prediction(object):
    __slots__ = [
        'input_data',
        'input_type',
        'output_data',
        'output_type',
        'context_data',
        'timestamp'
    ]

    DATA_TYPE_TABULAR = 1
    DATA_TYPE_TEXT = 2
    DATA_TYPE_IMAGE = 3

    _data_type_map = {
        'tabular': DATA_TYPE_TABULAR,
        'text': DATA_TYPE_TEXT,
        'image': DATA_TYPE_IMAGE
    }

    def __init__(
            self,
            input_data=None,
            input_type=None,
            output_data=None,
            output_type=None,
            context_data=None,
            timestamp=None):
        self.input_data = input_data
        self.input_type = input_type
        self.output_data = output_data
        self.output_type = output_type
        self.context_data = context_data
        self.timestamp = timestamp if timestamp is not None else _now()

    @staticmethod
    def data_type(data_type_name):
        return Prediction._data_type_map[data_type_name]


def _now():
    return int(time.time())



import unittest
import logging
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.windows import Metric

logger = logging.getLogger('graphsignal')


class WindowsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_update_percentile(self):
        data = [1, 2, 3, 4, 5] * 10
        data.append(100)
        data.append(1000)

        metric = Metric()
        for sample in data:
            metric.update_percentile(sample, 95)

        metric.finalize()

        self.assertEqual(
            metric.measurement,
            [5, 52]
        )

    def test_compute_histogram(self):
        data = [-0.22, -0.001, 0, 1, 1.1, 2, 2.00345]

        metric = Metric()
        metric.compute_histogram(data)

        self.assertEqual(
            metric.measurement,
            [0.1, -0.3, 1, -0.1, 1, 0.0, 1, 1.0, 1, 1.1, 1, 2.0, 2]
        )

    def test_compute_histogram_one_measurement(self):
        data = [212]

        metric = Metric()
        metric.compute_histogram(data)

        self.assertEqual(
            metric.measurement,
            [0, 212, 1]
        )

    def test_compute_histogram_same_measurement(self):
        data = [0.1111, 0.1111, 0.1111]

        metric = Metric()
        metric.compute_histogram(data)

        self.assertEqual(
            metric.measurement,
            [0, 0.1111, 3]
        )

    def test_compute_histogram_many_measurements(self):
        data = [i for i in range(15, 799)]

        metric = Metric()
        metric.compute_histogram(data)

        self.assertEqual(len(metric.measurement), 159)

    def test_compute_histogram_big_range(self):
        data = [-50032, 2000456]

        metric = Metric()
        metric.compute_histogram(data)

        self.assertEqual(
            metric.measurement,
            [100000, -100000, 1, 2000000, 1]
        )

    def test_compute_histogram_small_range(self):
        data = [-0.00000012, 0.000009]

        metric = Metric()
        metric.compute_histogram(data)

        self.assertEqual(
            metric.measurement,
            [0.0000001, -0.0000002, 1, 0.000009, 1]
        )

    def test_compute_categorical_histogram(self):
        data = ['a', 'b', 'b', 'c', 'd']

        metric = Metric()
        metric.compute_categorical_histogram(data)

        self.assertEqual(
            metric.measurement,
            [1, 43, 3, 53, 1, 97, 1]
        )



import os
from setuptools import setup, find_packages

def read(fname):
  return open(os.path.join(os.path.dirname(__file__), fname)).read()

version = {}
with open('graphsignal/version.py') as fp:
    exec(fp.read(), version)

setup(
  name = 'graphsignal',
  version = version['__version__'],
  description = 'Graphsignal Logger',
  long_description = read('README.md'),
  long_description_content_type = 'text/markdown',
  author = 'Graphsignal, Inc.',
  author_email = 'devops@graphsignal.ai',
  url = 'https://graphsignal.ai',
  license = 'BSD',
  keywords = [
    'machine learning',
    'deep learning',
    'data science',
    'MLOps',
    'machine learning devops',
    'machine learning monitoring',
    'ML monitoring',
    'AI monitoring',
    'pipeline monitoring',
    'model monitoring',
    'prediction monitoring',
    'data drift',
    'concept drift',
    'training-serving skew'
  ],
  classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Web Environment',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Topic :: Software Development',
    'Topic :: System :: Monitoring',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence'
  ],
  python_requires='>=3.5',
  install_requires=[
    'numpy',
    'pandas',
    'scipy',
    'scikit-learn'
  ],
  
  packages = find_packages(exclude=[
    '*.sh', '*_test.py', 'examples'])
)



import unittest
import logging
import time
from unittest.mock import patch, Mock
import pprint
import numpy as np

import graphsignal
from graphsignal.uploader import Uploader
logger = logging.getLogger('graphsignal')


class SessionsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        graphsignal.sessions.reset_all()
        graphsignal.configure(api_key='k1', debug_mode=True)

    def tearDown(self):
        graphsignal.sessions.reset_all()
        graphsignal.shutdown()

    @patch('platform.python_version', return_value='1.2.3')
    @patch('platform.system', return_value='Linux')
    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    @patch('graphsignal.statistics.compute_metrics', return_value=[])
    @patch('graphsignal.system.vm_size', return_value=1)
    @patch('graphsignal.system.vm_rss', return_value=1)
    @patch('time.time', return_value=1)
    def test_prediction_batched(self, mocked_time, mocked_vm_rss, mocked_vm_size, mocked_compute_metrics,
                                mocked_upload_window, mocked_flush, mocked_system, mocked_version):
        session = graphsignal.session('m1', 'd1')
        session.set_attribute(name='a1', value='v1')
        session.log_metric(name='m1', value=1)
        session.log_event(
            description='e1', attributes={
                'a1': 'v1'}, is_error=True)
        session.log_prediction(input_data=[[1, 2], [3, 4]], output_data=[5, 6])
        session._upload_window(force=True)

        mocked_upload_window.assert_called_once()

        call_args = mocked_upload_window.call_args[0][0]
        call_args['metrics'] = sorted(
            call_args['metrics'], key=lambda metric: metric['name'])
        #pp = pprint.PrettyPrinter()
        # pp.pprint(call_args)

        self.assertEqual(
            call_args,
            {'events': [{'attributes': [{'name': 'a1', 'value': 'v1'}],
                         'description': 'e1',
                         'name': 'user_defined_event',
                         'score': 1,
                         'timestamp': 1,
                         'type': 'error'}],
             'metrics': [{'dataset': 'user_defined',
                          'measurement': [1],
                          'name': 'm1',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': ''},
                         {'dataset': 'system',
                          'measurement': [1],
                          'name': 'prediction_count',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': ''},
                         {'dataset': 'system',
                          'measurement': [1],
                          'name': 'process_memory_usage',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': 'KB'},
                         {'dataset': 'system',
                          'measurement': [1],
                          'name': 'process_virtual_memory',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': 'KB'}],
             'model': {'attributes': [{'name': 'Python', 'value': '1.2.3'},
                                      {'name': 'OS', 'value': 'Linux'},
                                      {'name': 'a1', 'value': 'v1'}],
                       'name': 'm1',
                       'deployment': 'd1',
                       'timestamp': 1},
             'timestamp': 1}
        )

    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    def test_prediction_not_uploaded(self, mocked_upload_window, mocked_flush):
        session = graphsignal.session('m1')

        session.log_prediction(input_data=[[1, 2], [3, 4]])
        session.log_prediction(input_data=[[1, 2], [3, 4]])

        self.assertEqual(session._window_size, 4)
        mocked_upload_window.assert_not_called()

    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    def test_prediction_uploaded_max_window_duration(
            self, mocked_upload_window, mocked_flush):
        session = graphsignal.session('m1')

        session._window_start_time = session._window_start_time - \
            graphsignal.sessions.MAX_WINDOW_DURATION - 1
        session.log_prediction(input_data=[[1, 2], [3, 4]])

        mocked_upload_window.assert_called_once()

    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    def test_prediction_uploaded_min_window_size(
            self, mocked_upload_window, mocked_flush):
        session = graphsignal.session('m1')

        session._window_start_time = session._window_start_time - \
            graphsignal.sessions.MIN_WINDOW_DURATION - 1
        session.log_prediction(
            input_data=np.random.rand(
                graphsignal.sessions.MIN_WINDOW_SIZE, 2))

        mocked_upload_window.assert_called_once()

    @patch('platform.python_version', return_value='1.2.3')
    @patch('platform.system', return_value='Linux')
    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    @patch('graphsignal.system.vm_size', return_value=1)
    @patch('graphsignal.system.vm_rss', return_value=1)
    @patch('graphsignal.system.cpu_time', return_value=1)
    @patch('time.time', return_value=1)
    @patch('time.monotonic', return_value=1)
    def test_measure_latency_context(self, mocked_monotonic, mocked_time, mocked_cpu_time, mocked_vm_rss, mocked_vm_size,
                                     mocked_upload_window, mocked_flush, mocked_system, mocked_version):
        session = graphsignal.session('m1')

        try:
            with session.measure_latency():
                time.sleep(0.1)
                raise Exception('ex1')
        except BaseException:
            pass

        session._upload_window(force=True)

        window = mocked_upload_window.call_args[0][0]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(window)

        self.assertEqual(
            window,
            {'metrics': [{'dataset': 'system',
                          'measurement': [0, 1],
                          'name': 'prediction_latency_p50',
                          'timestamp': 1,
                          'type': 'statistic',
                          'unit': 'ms'},
                         {'dataset': 'system',
                          'measurement': [0, 1],
                          'name': 'prediction_cpu_time_p50',
                          'timestamp': 1,
                          'type': 'statistic',
                          'unit': 'ms'},
                         {'dataset': 'system',
                          'measurement': [0],
                          'name': 'prediction_count',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': ''},
                         {'dataset': 'system',
                          'measurement': [1],
                          'name': 'process_memory_usage',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': 'KB'},
                         {'dataset': 'system',
                          'measurement': [1],
                          'name': 'process_virtual_memory',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': 'KB'}],
             'model': {'attributes': [{'name': 'Python', 'value': '1.2.3'},
                                      {'name': 'OS', 'value': 'Linux'}],
                       'name': 'm1',
                       'timestamp': 1},
             'timestamp': 1}
        )

    @patch('platform.python_version', return_value='1.2.3')
    @patch('platform.system', return_value='Linux')
    @patch.object(Uploader, 'flush')
    @patch.object(Uploader, 'upload_window')
    @patch('graphsignal.system.vm_size', return_value=1)
    @patch('graphsignal.system.vm_rss', return_value=1)
    @patch('graphsignal.system.cpu_time', return_value=1)
    @patch('time.time', return_value=1)
    @patch('time.monotonic', return_value=1)
    def test_measure_latency_start_stop(
            self, mocked_monotonic, mocked_time, mocked_cpu_time, mocked_vm_rss, mocked_vm_size, mocked_upload_window, mocked_flush, mocked_system, mocked_version):
        session = graphsignal.session('m1')

        span = session.measure_latency()
        span.start()
        time.sleep(0.1)
        span.stop()

        session._upload_window(force=True)

        window = mocked_upload_window.call_args[0][0]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(window)

        self.assertEqual(
            window,
            {'metrics': [{'dataset': 'system',
                          'measurement': [0, 1],
                          'name': 'prediction_latency_p50',
                          'timestamp': 1,
                          'type': 'statistic',
                          'unit': 'ms'},
                         {'dataset': 'system',
                          'measurement': [0, 1],
                          'name': 'prediction_cpu_time_p50',
                          'timestamp': 1,
                          'type': 'statistic',
                          'unit': 'ms'},
                         {'dataset': 'system',
                          'measurement': [0],
                          'name': 'prediction_count',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': ''},
                         {'dataset': 'system',
                          'measurement': [1],
                          'name': 'process_memory_usage',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': 'KB'},
                         {'dataset': 'system',
                          'measurement': [1],
                          'name': 'process_virtual_memory',
                          'timestamp': 1,
                          'type': 'gauge',
                          'unit': 'KB'}],
             'model': {'attributes': [{'name': 'Python', 'value': '1.2.3'},
                                      {'name': 'OS', 'value': 'Linux'}],
                       'name': 'm1',
                       'timestamp': 1},
             'timestamp': 1}
        )



import time
import logging
import threading
import sys
import traceback
import platform

import graphsignal
from graphsignal import statistics
from graphsignal import system
from graphsignal.uploader import Uploader
from graphsignal.predictions import Prediction
from graphsignal.windows import Window, Model, Metric, Event
from graphsignal.spans import Span

logger = logging.getLogger('graphsignal')

MAX_MODEL_ATTRIBUTES = 10
MAX_METRICS = 50
MAX_EVENTS = 50

MIN_WINDOW_SIZE = 50
MIN_WINDOW_DURATION = 120
MAX_WINDOW_DURATION = 600

_session_index = {}
_session_index_lock = threading.Lock()


class Session(object):
    __slots__ = [
        '_model',
        '_model_attributes',
        '_metric_index',
        '_prediction_window',
        '_event_window',
        '_window_start_time',
        '_window_size',
        '_update_lock',
        '_is_updated',
        '_upload_timer'
    ]

    def __init__(self, model_name, deployment_name=None):
        self._model = Model(name=model_name, deployment=deployment_name)
        self._model_attributes = {}
        self._update_lock = threading.Lock()
        self._reset_window()
        self._add_system_attributes()

    def _reset_window(self):
        self._metric_index = {}
        self._prediction_window = []
        self._event_window = []
        self._window_start_time = time.time()
        self._window_size = 0
        self._is_updated = False

    def _set_updated(self):
        self._is_updated = True
        if self._upload_window():
            graphsignal._get_uploader().flush_in_thread()

    def set_attribute(self, name=None, value=None):
        '''
        Set model attributes. Only last set value is kept.

        Args:
            name (:obj:`str`):
                Model attribute name.
            value (:obj:`int` or :obj:`float`):
                Model attribute value.
        '''

        if not isinstance(name, str) or len(name) > 250:
            logger.error('invalid model attribute name format')
            return
        if not isinstance(value, (str, int, float)) or len(str(value)) > 2500:
            logger.error(
                'invalid model attribute value format for name: {0}'.format(name))
            return

        if len(self._model_attributes) >= MAX_MODEL_ATTRIBUTES:
            logger.error(
                'too many model attributes, max={0}'.format(MAX_MODEL_ATTRIBUTES))
            return

        self._model_attributes[name] = value

    def log_prediction(
            self,
            input_data=None,
            input_type='tabular',
            output_data=None,
            output_type='tabular',
            context_data=None,
            actual_timestamp=None):
        '''
        Log single or batch model prediction.

        See `Supported Data Formats <https://graphsignal.ai/docs/integrations/python/supported-data-formats/>`_
        for detailed description of data types and formats.

        Computed data statistics such as feature and class distributions are uploaded
        at certain intervals and on process exit. Additionally, some prediction samples
        and outliers may be uploaded. This can be disabled by setting ``log_instances`` option
        to ``False`` when configuring the logger.

        Please make sure to exclude or anonymize any personally identifiable information (PII).

        Args:
            input_data (:obj:`list` or :obj:`dict` or :obj:`numpy.ndarray` or :obj:`pandas.DataFrame`, optional):
                Input data representation.
            input_type (:obj:`str`, optional, default  is ``tabular``):
                Type of the provided input data.
            output_data (:obj:`list` or :obj:`dict` or :obj:`numpy.ndarray` or :obj:`pandas.DataFrame`, optional):
                Output data representation.
            output_type (:obj:`str`, optional, default  is ``tabular``):
                Type of the provided output data.
            context_data (:obj:`list` or :obj:`dict` or :obj:`numpy.ndarray` or :obj:`pandas.DataFrame`, optional):
                Context data, such as user ID, feature ID or any other prediction or data related information.
                It will be included in prediction samples and/or outliers. Context data is not monitored.
                The number of rows/instances in ``context_data`` should match ``input_data`` and/or ``output_data``.
            actual_timestamp (:obj:`int`, optional, default is current timestamp):
                Actual timestamp of the measurement, when different from current timestamp.
        '''

        if input_type != 'tabular':
            logger.error('Data type \'%s\' is not supported', input_type)
            return

        if output_type != 'tabular':
            logger.error('Data type \'%s\' is not supported', output_type)
            return

        self._window_size += max(
            statistics.estimate_size(input_data),
            statistics.estimate_size(output_type))

        with self._update_lock:
            self._prediction_window.append(Prediction(
                input_data=input_data,
                input_type=Prediction.data_type(input_type),
                output_data=output_data,
                output_type=Prediction.data_type(output_type),
                context_data=context_data,
                timestamp=actual_timestamp))

        self._set_updated()

    def log_metric(self, name=None, value=None, actual_timestamp=None):
        '''
        Logs metric value. The last logged value is used
        when metric is aggregated.

        Args:
            name (:obj:`str`):
                Metric name.
            value (:obj:`int` or :obj:`float`):
                Metric value.
            actual_timestamp (:obj:`int`, optional, default is current timestamp):
                Actual timestamp of the measurement, when different from current timestamp.
        '''

        if not isinstance(name, str) or len(name) > 250:
            logger.error('invalid metric name format')
            return
        if value is not None and not isinstance(
                value, (int, float)) or len(str(value)) > 2500:
            logger.error(
                'invalid metric value format for "{0}": {1}'.format(name, type(value)))
            return

        if len(self._metric_index) >= MAX_METRICS:
            logger.error('too many metrics, max={0}'.format(MAX_METRICS))
            return

        with self._update_lock:
            metric = None
            if name in self._metric_index:
                metric = self._metric_index[name]
            else:
                metric = self._metric_index[name] = Metric(
                    dataset=Metric.DATASET_USER_DEFINED,
                    name=name,
                    timestamp=actual_timestamp)
            metric.set_gauge(value)

        self._set_updated()

    def log_event(
            self,
            description=None,
            attributes=None,
            is_error=False,
            exc_info=None,
            actual_timestamp=None):
        '''
        Log arbitrary event or exception.

        Args:
            description (:obj:`str`):
                Event description.
            attributes (:obj:`dict`, optional):
                Event attributes.
            is_error (:obj:`bool`, optional):
                Set error type.
            exc_info (:obj:`bool` or :obj:`tuple`, optional):
                If the event should represents an exception, pass exception tuple
                or ``True`` to automatically read it from ``sys.exc_info()``.
            actual_timestamp (:obj:`int`, optional, default is current timestamp):
                Actual timestamp of the measurement, when different from current timestamp.
        '''

        if not description or not isinstance(
                description, str) or len(description) > 250:
            logger.error('invalid format for description')
            return

        if attributes is not None:
            if isinstance(attributes, dict):
                for name, value in attributes.items():
                    if not isinstance(name, str) or len(name) > 250:
                        logger.error('invalid attribute name format')
                        return
                    if not isinstance(value, (str, int, float)):
                        logger.error(
                            'invalid attribute value format for attribute name {0}'.format(name))
                        return
            else:
                logger.error('invalid attributes format, expecting dict')
                return

        if len(self._event_window) >= MAX_EVENTS:
            logger.error('too many events, max={0}'.format(MAX_EVENTS))
            return

        type_name = Event.TYPE_INFO
        if is_error or exc_info:
            type_name = Event.TYPE_ERROR

        if exc_info:
            event_name = 'exception'
            if exc_info == True:
                exc_info = sys.exc_info()
            if len(
                    exc_info) == 3 and exc_info[0] and exc_info[1] and exc_info[2]:
                attributes['error_message'] = traceback.format_exception_only(
                    exc_info[0], exc_info[1])
                attributes['stack_trace'] = traceback.format_tb(exc_info[2])
        else:
            event_name = 'user_defined_event'

        with self._update_lock:
            event = Event(
                type=type_name,
                name=event_name,
                description=description,
                timestamp=actual_timestamp)
            for name, value in attributes.items():
                event.add_attribute(name, value)
            self._event_window.append(event)

        self._set_updated()

    def measure_latency(self):
        '''
        Measure prediction/inference call latency.

        When used as a context manager, will also record any exceptions.

        Otherwise, ``start()`` and ``stop()`` methods should be called
        on returned ``Span`` object to start and stop latency measurement.

        Returns:
            :obj:`Span` - span object for measuring latency.
        '''

        return Span(self)

    def _upload_window(self, force=False):
        if not self._is_updated:
            return False

        # check if current window should be uploaded
        if not force:
            window_duration = time.time() - self._window_start_time
            if window_duration < MIN_WINDOW_DURATION:
                return False
            if (self._window_size < MIN_WINDOW_SIZE and
                    window_duration < MAX_WINDOW_DURATION):
                return False

        # reset
        with self._update_lock:
            metric_index = self._metric_index
            prediction_window = self._prediction_window
            events_window = self._event_window
            self._reset_window()

        # initialize window object
        window = Window()

        window.model = self._model
        if self._model_attributes is not None:
            for name, value in self._model_attributes.items():
                window.model.add_attribute(name, value)

        # add user defined metrics
        user_metrics = metric_index.values()
        for metric in user_metrics:
            window.add_metric(metric)

        # add prediction count metric
        last_timestamp = max([p.timestamp for p in prediction_window if p]) if len(prediction_window) > 0 else None
        prediction_count_metric = Metric(
            dataset=Metric.DATASET_SYSTEM,
            name='prediction_count',
            timestamp=last_timestamp)
        prediction_count_metric.set_gauge(len(prediction_window))
        window.add_metric(prediction_count_metric)

        # add computed data metrics
        try:
            data_metrics, data_samples = statistics.compute_metrics(
                prediction_window)
            if data_metrics is not None and len(data_metrics) > 0:
                for metric in data_metrics:
                    window.add_metric(metric)
            if data_samples is not None and len(data_samples) > 0:
                for sample in data_samples:
                    window.add_sample(sample)
        except Exception:
            logger.error(
                'Unable to compute data statistics', exc_info=True)

        # add system metrics
        if graphsignal._get_config().log_system_metrics:
            vm_rss = system.vm_rss()
            if vm_rss is not None:
                vm_rss_metric = Metric(
                    dataset=Metric.DATASET_SYSTEM,
                    name='process_memory_usage')
                vm_rss_metric.set_gauge(vm_rss, unit=Metric.UNIT_KILOBYTE)
                window.add_metric(vm_rss_metric)

            vm_size = system.vm_size()
            if vm_size is not None:
                vm_size_metric = Metric(
                    dataset=Metric.DATASET_SYSTEM,
                    name='process_virtual_memory')
                vm_size_metric.set_gauge(vm_size, unit=Metric.UNIT_KILOBYTE)
                window.add_metric(vm_size_metric)

        # finalize metrics
        for metric in window.metrics:
            metric.finalize()

        # add events
        for event in events_window:
            window.add_event(event)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Uploading window:')
            logger.debug(window)

        graphsignal._get_uploader().upload_window(window.to_dict())
        return True

    def _add_system_attributes(self):
        self._model.add_attribute('Python', platform.python_version())
        self._model.add_attribute('OS', platform.system())


def get_session(model_name, deployment_name=None):
    if not isinstance(model_name, str) or len(model_name) > 250:
        raise ValueError('invalid model name format')

    if deployment_name is not None and (not isinstance(deployment_name, str) or len(deployment_name) > 250):
        raise ValueError('invalid deployment_name format')

    model_key = model_name
    if deployment_name is not None:
        model_key += model_name

    with _session_index_lock:
        if model_key in _session_index:
            return _session_index[model_key]
        else:
            sess = Session(model_name, deployment_name)
            _session_index[model_key] = sess
            return sess


def reset_all():
    with _session_index_lock:
        _session_index.clear()


def upload_all(force=False):
    session_list = None
    with _session_index_lock:
        session_list = _session_index.values()

    uploaded = False
    for session in session_list:
        if session._upload_window(force=force):
            uploaded = True

    return uploaded



import os
import time
import json
import gzip
import sys
import threading
import base64
import logging
from io import BytesIO
from urllib.request import urlopen
from urllib.request import Request
from urllib.parse import urlencode
from urllib.error import URLError
from urllib.error import HTTPError

import graphsignal

logger = logging.getLogger('graphsignal')


class Uploader(object):
    MAX_BUFFER_SIZE = 2500

    def __init__(self):
        self.collector_url = 'https://log-api.graphsignal.ai'
        self.buffer = {}
        self.buffer_lock = threading.Lock()
        self.flush_lock = threading.Lock()

    def configure(self):
        if 'GRAPHSIGNAL_LOG_API_URL' in os.environ:
            self.collector_url = os.environ['GRAPHSIGNAL_LOG_API_URL']

    def clear(self):
        with self.buffer_lock:
            self.buffer = {}

    def upload_window(self, window):
        self.upload_json('windows', [window])

    def upload_json(self, endpoint, data):
        with self.buffer_lock:
            if endpoint not in self.buffer:
                self.buffer[endpoint] = []
            self.buffer[endpoint].extend(data)
            if len(self.buffer[endpoint]) > self.MAX_BUFFER_SIZE:
                self.buffer[endpoint] = self.buffer[endpoint][-self.MAX_BUFFER_SIZE:]

    def flush_in_thread(self):
        threading.Thread(target=self.flush).start()

    def flush(self):
        with self.flush_lock:
            for endpoint, outgoing in self.buffer.items():
                with self.buffer_lock:
                    if len(outgoing) == 0:
                        continue
                    self.buffer[endpoint] = []
                try:
                    upload_start = time.time()
                    self.post_json(endpoint, outgoing)
                    logger.debug(
                        'Upload to %s/%s took %.3f sec', self.collector_url, endpoint, time.time() - upload_start)
                except URLError:
                    logger.debug(
                        'Failed uploading to %s/%s', self.collector_url, endpoint, exc_info=True)
                    with self.buffer_lock:
                        self.buffer[endpoint][:0] = outgoing
                except Exception:
                    logger.debug(
                        'Exception when uploading to %s/%s', self.collector_url, endpoint, exc_info=True)

    def post_json(self, endpoint, data):
        logger.debug('Posting data to %s/%s: %s',
                     self.collector_url, endpoint, data)

        api_key_64 = _base64_encode(
            graphsignal._get_config().api_key + ':').replace('\n', '')
        headers = {
            'Accept-Encoding': 'gzip',
            'Authorization': "Basic %s" % api_key_64,
            'Content-Type': 'application/json',
            'Content-Encoding': 'gzip'
        }

        gzip_out = BytesIO()
        with gzip.GzipFile(fileobj=gzip_out, mode="w") as out_file:
            out_file.write(json.dumps(data).encode('utf-8'))
            out_file.close()

        gzip_out_val = gzip_out.getvalue()
        if isinstance(gzip_out_val, str):
            data_gzip = bytearray(gzip_out.getvalue())
        else:
            data_gzip = gzip_out.getvalue()
        request = Request(
            url=self.collector_url + '/' + endpoint,
            data=data_gzip,
            headers=headers)

        try:
            resp = urlopen(request, timeout=10)
            result_data = resp.read()
            if resp.info():
                content_type = resp.info().get('Content-Encoding')
                if content_type == 'gzip':
                    result_data = gzip.GzipFile(
                        '', 'r', 0, BytesIO(result_data)).read()
            resp.close()
            return json.loads(result_data.decode('utf-8'))
        except HTTPError as herr:
            logger.debug(
                'Failed uploading to %s/%s', self.collector_url, endpoint)
            logger.debug('Error message from server: %s',
                         herr.read().decode('utf-8'))
            return {}


def _timestamp():
    return int(time.time())


def _base64_encode(s):
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')



import time
import unittest
import logging
from unittest.mock import patch, Mock
import numpy as np
import pandas as pd
import pprint

import graphsignal
from graphsignal import statistics as ds
from graphsignal.windows import Metric
from graphsignal.predictions import Prediction

logger = logging.getLogger('graphsignal')


class StatisticsTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        graphsignal.configure(api_key='k1', debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_data_size(self):
        self.assertEqual(ds.estimate_size([[1], [2]]), 2)
        self.assertEqual(ds.estimate_size(np.array([[1], [2]])), 2)
        self.assertEqual(ds.estimate_size({'a': [1, 2, 3], 'b': [4, 5, 6]}), 3)
        self.assertEqual(ds.estimate_size(
            {'a': np.asarray([1, 2, 3]), 'b': np.asarray([4, 5, 6])}), 3)
        self.assertEqual(ds.estimate_size(pd.DataFrame(data=[[1], [2]])), 2)

    def test_convert_to_2d_dict_of_lists(self):
        d = {'c1': [1, 2], 'c2': [3.1, 4.1]}
        data2d = ds._convert_to_2d(d)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1, 3.1], [2, 4.1]])))
        self.assertEqual(data2d.shape[0], 2)
        self.assertEqual(data2d.columns.values.tolist(), ['c1', 'c2'])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()], [
                         'int64', 'float64'])

    def test_convert_to_2d_list_of_dicts(self):
        d = [{'c1': 1, 'c2': 2.1}, {'c1': 3, 'c2': 4.1}]
        data2d = ds._convert_to_2d(d)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1, 2.1], [3, 4.1]])))
        self.assertEqual(data2d.shape[0], 2)
        self.assertEqual(data2d.columns.values.tolist(), ['c1', 'c2'])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()], [
                         'int64', 'float64'])

    def test_convert_to_2d_list_of_nested_dicts(self):
        d = [{'c1': 1, 'c2': {'c3': 2.1}}, {'c1': 3, 'c2': {'c3': 4.1}}]
        data2d = ds._convert_to_2d(d)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1], [3]])))
        self.assertEqual(data2d.shape[0], 2)
        self.assertEqual(data2d.columns.values.tolist(), ['c1'])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()], ['int64'])

    def test_convert_to_2d_list_of_ndarray(self):
        d = [np.asarray([1, 3.1]), np.asarray([2, 4.1])]
        data2d = ds._convert_to_2d(d)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1, 3.1], [2, 4.1]])))
        self.assertEqual(data2d.shape[0], 2)
        self.assertEqual(data2d.columns.values.tolist(), [0, 1])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()],
                         ['float64', 'float64'])

    def test_convert_to_2d_dict_of_ndarray(self):
        d = {'c1': np.asarray([1, 2]), 'c2': np.asarray([3.1, 4.1])}
        data2d = ds._convert_to_2d(d)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1, 3.1], [2, 4.1]])))
        self.assertEqual(data2d.shape[0], 2)
        self.assertEqual(data2d.columns.values.tolist(), ['c1', 'c2'])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()], [
                         'int64', 'float64'])

    def test_convert_to_2d_dataframe_with_columns(self):
        df = pd.DataFrame(data={'c1': [1, 2], 'c2': [3.1, 4.1]})
        data2d = ds._convert_to_2d(df)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1, 3.1], [2, 4.1]])))
        self.assertEqual(data2d.shape[0], 2)
        self.assertEqual(data2d.columns.values.tolist(), ['c1', 'c2'])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()], [
                         'int64', 'float64'])

    def test_convert_to_2d_dataframe_wo_columns(self):
        df = pd.DataFrame(data=[[1, 3.1], [2, 4.1]])
        data2d = ds._convert_to_2d(df)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1, 3.1], [2, 4.1]])))
        self.assertEqual(data2d.shape[0], 2)
        self.assertEqual(data2d.columns.values.tolist(), [0, 1])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()], [
                         'int64', 'float64'])

    def test_convert_to_2d_ndarray(self):
        d = np.asarray([[1, 3.1], [2, 4.1]])
        data2d = ds._convert_to_2d(d)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(
            data2d.to_numpy(), np.asarray([[1, 3.1], [2, 4.1]])))
        self.assertEqual(data2d.columns.values.tolist(), [0, 1])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()],
                         ['float64', 'float64'])

    def test_convert_to_2d_ndarray_3d(self):
        d = np.full((2, 5, 3), [1, 2, 3.3])
        data2d = ds._convert_to_2d(d)
        self.assertIsNotNone(data2d)
        self.assertTrue(np.array_equal(data2d.to_numpy(), np.asarray([
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3],
            [1, 2, 3.3]
        ])))
        self.assertEqual(data2d.shape[0], 10)
        self.assertEqual(data2d.columns.values.tolist(), [0, 1, 2])
        self.assertEqual([str(t) for t in data2d.dtypes.tolist()], [
                         'float64', 'float64', 'float64'])

    @patch('time.time', return_value=1)
    def test_compute_metrics_empty(self, mocked_time):
        metrics, samples = ds.compute_metrics(
            [Prediction(input_data=[], output_data=[])])
        metric_dicts = [metric.to_dict() for metric in metrics]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(metric_dicts)
        self.assertEqual(metric_dicts,
                         []
                         )
        self.assertEqual(samples, [])

    @patch('time.time', return_value=1)
    def test_compute_metrics_simple(self, mocked_time):
        input_data = pd.DataFrame(
            data={'f1': [1, 1, 2, 0], 'f2': [3.5, 4.5, 5.5, 0]})
        output_data = pd.DataFrame(
            data={
                'c1': [
                    0.1, 0.2, 0.1, 0.4], 'c2': [
                    5, 10, 7, 12]})
        metrics, _ = ds.compute_metrics(
            [Prediction(input_data=input_data, output_data=output_data)])
        metric_dicts = [metric.to_dict() for metric in metrics]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(metric_dicts)
        self.assertEqual(
            metric_dicts,
            [{'dataset': 'input',
              'measurement': [4, 4],
              'name': 'instance_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'measurement': [2, 4],
              'name': 'dimension_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0.0, 4],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [25.0, 4],
              'name': 'zero_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [3, 4],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [1.0, 4],
              'name': 'mean',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0.816496580927726, 4],
              'name': 'standard_deviation',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0.0, 4],
              'name': 'skewness',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [1.5, 4],
              'name': 'kurtosis',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0.1, 0.0, 1, 1.0, 2, 2.0, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [0.0, 4],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [25.0, 4],
              'name': 'zero_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [4, 4],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [3.375, 4],
              'name': 'mean',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [2.3935677693908453, 4],
              'name': 'standard_deviation',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [-1.333118339791635, 4],
              'name': 'skewness',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [1.9096859504132233, 4],
              'name': 'kurtosis',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f2',
              'measurement': [0.1, 0.0, 1, 3.5, 1, 4.5, 1, 5.5, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''},
             {'dataset': 'output',
              'measurement': [4, 4],
              'name': 'instance_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'measurement': [2, 4],
              'name': 'dimension_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.0, 4],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.0, 4],
              'name': 'zero_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [3, 4],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.2, 4],
              'name': 'mean',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.14142135623730953, 4],
              'name': 'standard_deviation',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [1.4142135623730951, 4],
              'name': 'skewness',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [1.5000000000000018, 4],
              'name': 'kurtosis',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.01, 0.1, 2, 0.2, 1, 0.4, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [0.0, 4],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [0.0, 4],
              'name': 'zero_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [4, 4],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [8.5, 4],
              'name': 'mean',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [3.1091263510296048, 4],
              'name': 'standard_deviation',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [0.0, 4],
              'name': 'skewness',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [-2.4328180737217604, 4],
              'name': 'kurtosis',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c2',
              'measurement': [0.1, 5.0, 1, 7.0, 1, 10.0, 1, 12.0, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''}]
        )

    @patch('time.time', return_value=1)
    def test_compute_metrics_nan_inf(self, mocked_time):
        input_data = pd.DataFrame(data={'f1': [1, float('nan')]})
        output_data = pd.DataFrame(data={'c1': [float('inf'), 2]})
        metrics, _ = ds.compute_metrics(
            [Prediction(input_data=input_data, output_data=output_data)])
        metric_dicts = [metric.to_dict() for metric in metrics]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(metric_dicts)
        self.assertEqual(
            metric_dicts,
            [{'dataset': 'input',
              'measurement': [2, 2],
              'name': 'instance_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'measurement': [1, 2],
              'name': 'dimension_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [50.0, 2],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0.0, 2],
              'name': 'zero_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [2, 2],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [1.0, 2],
              'name': 'mean',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0.0, 2],
              'name': 'standard_deviation',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0.0, 2],
              'name': 'skewness',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [-3.0, 2],
              'name': 'kurtosis',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': 'f1',
              'measurement': [0, 1.0, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''},
             {'dataset': 'output',
              'measurement': [2, 2],
              'name': 'instance_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'measurement': [1, 2],
              'name': 'dimension_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [50.0, 2],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.0, 2],
              'name': 'zero_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [2, 2],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [2.0, 2],
              'name': 'mean',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.0, 2],
              'name': 'standard_deviation',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0.0, 2],
              'name': 'skewness',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [-3.0, 2],
              'name': 'kurtosis',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': 'c1',
              'measurement': [0, 2.0, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''}]
        )

    @patch('time.time', return_value=1)
    def test_compute_metrics_categorical(self, mocked_time):
        input_data = pd.DataFrame(
            data=[['a', 1], ['b', 2], ['b', 2], ['c', 3]])
        output_data = pd.DataFrame(data=['d', 'e', 'e', 'f', 'f', None, ''])
        metrics, _ = ds.compute_metrics(
            [Prediction(input_data=input_data, output_data=output_data)])
        metric_dicts = [metric.to_dict() for metric in metrics]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(metric_dicts)
        self.assertEqual(
            metric_dicts,
            [{'dataset': 'input',
              'measurement': [4, 4],
              'name': 'instance_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'measurement': [2, 4],
              'name': 'dimension_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '0',
              'measurement': [0.0, 4],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': '0',
              'measurement': [0.0, 4],
              'name': 'empty_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': '0',
              'measurement': [3, 4],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '0',
              'measurement': [1, 43, 3, 97, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [0.0, 4],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [0.0, 4],
              'name': 'zero_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [3, 4],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [2.0, 4],
              'name': 'mean',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [0.816496580927726, 4],
              'name': 'standard_deviation',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [0.0, 4],
              'name': 'skewness',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [1.5, 4],
              'name': 'kurtosis',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'input',
              'dimension': '1',
              'measurement': [0.1, 1.0, 1, 2.0, 2, 3.0, 1],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''},
             {'dataset': 'output',
              'measurement': [7, 7],
              'name': 'instance_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'measurement': [1, 7],
              'name': 'dimension_count',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': '0',
              'measurement': [14.285714285714285, 7],
              'name': 'missing_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': '0',
              'measurement': [14.285714285714285, 7],
              'name': 'empty_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': '%'},
             {'dataset': 'output',
              'dimension': '0',
              'measurement': [4, 7],
              'name': 'unique_values',
              'timestamp': 1,
              'type': 'statistic',
              'unit': ''},
             {'dataset': 'output',
              'dimension': '0',
              'measurement': [1, 53, 1, 59, 2, 66, 3],
              'name': 'distribution',
              'timestamp': 1,
              'type': 'histogram',
              'unit': ''}]
        )

    @patch('graphsignal.statistics._random_index', return_value=[10, 20, 30])
    @patch('time.time', return_value=1)
    def test_compute_metrics_samples(self, mocked_time, mocked_random_index):
        r = []
        r.extend(list(range(100)))
        r.append(1001)
        r.extend(list(range(100)))
        r.append(1002)
        c = ['abc' + str(i) for i in r]
        input_data = pd.DataFrame(data={'f1': r, 'f2': r})
        output_data = pd.DataFrame(data={'c1': r})
        context_data = pd.DataFrame(data={'ctx1': c})
        _, samples = ds.compute_metrics([Prediction(
            input_data=input_data, output_data=output_data, context_data=context_data)])
        samples_dict = [sample.to_dict() for sample in samples]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(samples_dict)
        self.assertEqual(
            samples_dict,
            [{'name': 'random_sample',
              'segments': [{'data': 'f1,f2\n10,10\n20,20\n30,30',
                            'dataset': 'input',
                            'format': 'csv'},
                           {'data': 'c1\n10\n20\n30',
                            'dataset': 'output',
                            'format': 'csv'},
                           {'data': 'ctx1,prediction_timestamp\nabc10,1\nabc20,1\nabc30,1',
                            'dataset': 'context',
                            'format': 'csv'}],
              'size': 3,
              'timestamp': 1},
             {'name': 'input_outliers',
              'segments': [{'data': 'f1,f2\n1002,1002\n1001,1001',
                            'dataset': 'input',
                            'format': 'csv'},
                           {'data': 'c1\n1002\n1001',
                            'dataset': 'output',
                            'format': 'csv'},
                           {'data': 'ctx1,prediction_timestamp\nabc1002,1\nabc1001,1',
                            'dataset': 'context',
                            'format': 'csv'}],
              'size': 2,
              'timestamp': 1},
             {'name': 'output_outliers',
              'segments': [{'data': 'c1\n1002\n1001',
                            'dataset': 'output',
                            'format': 'csv'},
                           {'data': 'f1,f2\n1002,1002\n1001,1001',
                            'dataset': 'input',
                            'format': 'csv'},
                           {'data': 'ctx1,prediction_timestamp\nabc1002,1\nabc1001,1',
                            'dataset': 'context',
                            'format': 'csv'}],
              'size': 2,
              'timestamp': 1}]
        )

    @patch('time.time', return_value=1)
    def test_compute_metrics_sample_single(self, mocked_time):
        input_data = pd.DataFrame(data={'f1': [1], 'f2': [2]})
        output_data = pd.DataFrame(data={'c1': [3]})
        _, samples = ds.compute_metrics(
            [Prediction(input_data=input_data, output_data=output_data)])
        sample_dict = [sample.to_dict() for sample in samples]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(sample_dict)
        self.assertEqual(
            sample_dict,
            [{'name': 'random_sample',
              'segments': [{'data': 'f1,f2\n1,2', 'dataset': 'input', 'format': 'csv'},
                           {'data': 'c1\n3', 'dataset': 'output', 'format': 'csv'},
                           {'data': 'prediction_timestamp\n1',
                            'dataset': 'context',
                            'format': 'csv'}],
              'size': 1,
              'timestamp': 1}]
        )

    @patch('graphsignal.statistics._random_index', return_value=[0])
    @patch('time.time', return_value=1)
    def test_compute_metrics_sample_no_outliers(
            self, mocked_time, mocked_random_index):
        r = [1] * 100
        input_data = pd.DataFrame(data={'f1': r, 'f2': r})
        output_data = pd.DataFrame(data={'c1': r})
        _, samples = ds.compute_metrics(
            [Prediction(input_data=input_data, output_data=output_data)])
        sample_dict = [sample.to_dict() for sample in samples]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(sample_dict)
        self.assertEqual(
            sample_dict,
            [{'name': 'random_sample',
              'segments': [{'data': 'f1,f2\n1,1', 'dataset': 'input', 'format': 'csv'},
                           {'data': 'c1\n1', 'dataset': 'output', 'format': 'csv'},
                           {'data': 'prediction_timestamp\n1',
                            'dataset': 'context',
                            'format': 'csv'}],
              'size': 1,
              'timestamp': 1}]
        )

    @patch('graphsignal.statistics._random_index', return_value=[0])
    @patch('time.time', return_value=1)
    def test_compute_metrics_sample_missing_values(
            self, mocked_time, mocked_random_index):
        r = [[None, np.nan, np.inf, float('nan'), float('inf')]] * 100
        input_data = pd.DataFrame(data=r)
        output_data = pd.DataFrame(data=r)
        _, samples = ds.compute_metrics(
            [Prediction(input_data=input_data, output_data=output_data)])
        sample_dict = [sample.to_dict() for sample in samples]
        #pp = pprint.PrettyPrinter()
        # pp.pprint(sample_dict)
        self.assertEqual(
            sample_dict,
            [{'name': 'random_sample',
              'segments': [{'data': '0,1,2,3,4\nNone,nan,inf,nan,inf',
                            'dataset': 'input',
                            'format': 'csv'},
                           {'data': '0,1,2,3,4\nNone,nan,inf,nan,inf',
                            'dataset': 'output',
                            'format': 'csv'},
                           {'data': 'prediction_timestamp\n1',
                            'dataset': 'context',
                            'format': 'csv'}],
              'size': 1,
              'timestamp': 1}]
        )
